/*************************************************************
Generated for IBM Rational SDL & TTCN Suite 6.3.0.3.23566 at Sat Nov 17 11:52:49 AM CET 2012
Licensed Materials - Property of IBM 
(c) Copyright IBM Corporation 1999, 2012.  All Rights Reserved.

U.S. Government Users Restricted Rights:  
Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
*************************************************************/
/*****************************************************************************

  C-code macros for RTOS-integrations

  To be used with Cadvanced SDL to C Compiler 4.4

*****************************************************************************/

#include "scttypes.h"

#ifdef RTOSTEST
#include "rebootLib.h"
#include "syslib.h"
#endif

#ifdef X_SIGNAL_PRIO
/*---+--------------------------------------------------------------
     Signal priority upgrade (fab)
     Association object among threads and intermediate queues
-------------------------------------------------------------------*/
TAssociationObj AssociationObj;

/*---+--------------------------------------------------------------
     Signal priority upgrade (fab)
	 Redefinition of msgQReceive, msgQNumMsgs, taskSpawn, taskDelete
-------------------------------------------------------------------*/

int TmsgQReceive (MSG_Q_ID msgQId, char* buffer, UINT maxNBytes, int timeout) {
  TIQElem Elem;
  TIntermediateQueue* Q;
  /* Access the queue */
  Q = TAOGetQueue (&AssociationObj, taskIdSelf ());
  /* Extract all messages from the queue */
  while (msgQNumMsgs (msgQId) != 0) {
    /* using caller space prevents us from allocating the stack in depth */
    Elem.Len = msgQReceive (msgQId, buffer, maxNBytes, NO_WAIT);
    Elem.Data = xAlloc (Elem.Len);
	memcpy (Elem.Data, buffer, Elem.Len);
	Elem.Priority = ((xSignalHeader) buffer)->Priority;
    TIQAdd (Q, &Elem); /* add them to the intermediate queue */
  }
  /* Return the first message from the queue or wait another message if empty */
  if (!TIQIsEmpty (Q)) {
    TIQExtract (Q, &Elem);
    memcpy (buffer, Elem.Data, Elem.Len);
	xFree (&Elem.Data);
	return Elem.Len;
  } 
  /* if both system and intermediate queues are empty, can avoid to insert and extract
     next message from the queue */
  return msgQReceive (msgQId, buffer, maxNBytes, timeout);
}

int TmsgQNumMsgs (MSG_Q_ID msgQId) {
  /* should return the sum of the emulated function plus the number 
     of messages in the intermediate queue */
  TIntermediateQueue* Q;
	int Ret;
  /* Access the queue */
  Q = TAOGetQueue (&AssociationObj, taskIdSelf ());
  Ret = msgQNumMsgs (msgQId) + TIQCountMsg (Q);
	return Ret;
}

int TtaskSpawn (char* name,	int priority, int options, int stackSize, FUNCPTR entryPt,
	int arg1, int arg2,	int arg3, int arg4,	int arg5, int arg6,	int arg7, int arg8,
	int arg9, int arg10) {
  int Ret;
  Ret = taskSpawn (name, priority, options, stackSize, entryPt,
	               arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10);
  TAORegisterTask (&AssociationObj, Ret);
  return Ret;
}

STATUS TtaskDelete (int tid) {
  TAOFreeTask (&AssociationObj, taskIdSelf ());
  return taskDelete (tid);
}

/*---+--------------------------------------------------------------
     Signal priority upgrade (fab)
	 Intermediate queue implementation
-------------------------------------------------------------------*/

void TIQInit (TIntermediateQueue* Q) {Q->Head = NULL;}

void TIQAdd (TIntermediateQueue* Q, TIQElem* S) {
  /* We navigate the queue from head to tail. The incoming signal should be added 
     before the first signal having less priority that the new one (i.e., having a
	 greater priority index) */
  TIQElem* NewEl;
  TIQElem* P;
  TIQElem* SucOfP;

  /* Copying the element */
  NewEl = (TIQElem*) xAlloc (sizeof (*S));
  memcpy (NewEl, S, sizeof (*NewEl));
  /* Navigating the list and inserting the element */
  P = Q->Head;
  SucOfP = NULL;
  while (P) { /* exit loop when P points the element where to insert
                                        the new one, between *P and its Suc */
	if (!P) break; /* go to add at tail */
	else if (P->Priority > NewEl->Priority) break; /* go to add inside */
    SucOfP = P;
   	P = P->Pre;
  }

  NewEl->Pre = P; /* can be NULL, but it's OK */
  if (SucOfP) SucOfP->Pre = NewEl; /* handle classical list */
  else Q->Head = NewEl;
}

int  TIQIsEmpty (TIntermediateQueue* Q) {
  return Q->Head == NULL;
}

void TIQExtract (TIntermediateQueue* Q, TIQElem* S) {
  if (!TIQIsEmpty (Q)) { /* melius est abundare quam deficere */
    memcpy (S, Q->Head, sizeof (*S));
	xFree (&Q->Head);
	Q->Head = S->Pre; /* can be NULL, it's OK */
	S->Pre = NULL; /* never leave unwanted accesses */
  }
}

void TIQForceFlush (TIntermediateQueue* Q) {
  TIQElem S;
  while (!TIQIsEmpty (Q)) {
    TIQExtract (Q, &S);
    if (S.Data) xFree (&S.Data);
  }
}

int TIQCountMsg (TIntermediateQueue* Q) {
  int Ret = 0;
  TIQElem* P;
  P = Q->Head;
  while (P) {
    Ret ++;
   	P = P->Pre;
  }
  return Ret;    
}

/*---+--------------------------------------------------------------
     Signal priority upgrade (fab)
	 Thread-Intermediate queue association object implementation
-------------------------------------------------------------------*/

void TAOInit (TAssociationObj* A) {
  int I;
  for (I = 0; I < ASS_ELEMENTS; I ++) {
    A->AssArray [I].Free = 1;
  }
	TAOCreateSemaphore (A);
}

void TAORegisterTask (TAssociationObj* A, TTaskId Id) {
  int I;
  /* Integrity protection */
  TAOWaitSemaphore (A);
  /* Find a free location */
  for (I = 0; I < ASS_ELEMENTS; I ++) {
    if (A->AssArray [I].Free) break;
  }
  if (I >= ASS_ELEMENTS) {
    XOS_ERROR ("TAORegisterTask", "List Is Full", 0);
	exit (1);
  } else {
    A->AssArray [I].Free = 0;
    A->AssArray [I].TaskId = Id;
    TIQInit (&(A->AssArray [I].Queue));
  }
  /* Integrity protection */
  TAOGiveSemaphore (A);
}

/* Free a task */
void TAOFreeTask (TAssociationObj* A, TTaskId Id) {
  int I;
  /* Integrity protection */
  TAOWaitSemaphore (A);
  /* Find the location */
  for (I = 0; I < ASS_ELEMENTS; I ++) {
    if (A->AssArray [I].TaskId == Id && !(A->AssArray [I].Free)) break;
  }
  if (I >= ASS_ELEMENTS) {
    XOS_ERROR ("TAOFreeTask", "Task Not Found", Id);
	exit (1);
  } else {
    A->AssArray [I].Free = 1;
    TIQForceFlush (&(A->AssArray [I].Queue));
  }
  /* Integrity protection */
  TAOGiveSemaphore (A);
}

/* Get intermediate queue access for a given task */
TIntermediateQueue* TAOGetQueue (TAssociationObj* A, TTaskId Id) {
  int I;
  TIntermediateQueue* Ret;
  /* Integrity protection */
  TAOWaitSemaphore (A);
  /* Find the location */
  for (I = 0; I < ASS_ELEMENTS; I ++) {
    if (A->AssArray [I].TaskId == Id && !(A->AssArray [I].Free)) break;
  }
  if (I >= ASS_ELEMENTS) {
    XOS_ERROR ("TAOGetQueue", "Task Not Found", Id);
  	exit (1);
  }
  Ret = &(A->AssArray [I].Queue);
  /* Integrity protection */
  TAOGiveSemaphore (A);
  return Ret;
}

void TAOCreateSemaphore (TAssociationObj* A) {
  A->Semaphore = semCCreate (SEM_Q_FIFO, 1);
}

void TAOWaitSemaphore (TAssociationObj* A) {
  semTake (A->Semaphore, WAIT_FOREVER);
}

void TAOGiveSemaphore (TAssociationObj* A) {
  semGive (A->Semaphore);
}

/*---+--------------------------------------------------------------
     Instance number check (fab)
	 Function which counts active instances given a symtab  node
-------------------------------------------------------------------*/

#ifdef X_INSTANCE_CHECK
int GetActiveInstances (xPrsIdNode Node) {
  int Ret = 0;
  xEPrsNode P;
  RTOSALLOCPRSSEMTAKE
  P = *Node->ActivePrsList;
  while (P) {
    Ret ++;
	P = P->NextPrs;
  }
  RTOSALLOCPRSSEMGIVE
  /* printf ("Returning %d\n", Ret); */
  return Ret;
}
#endif

#endif
 /* X_SIGNAL_PRIO */


#ifdef RTOSTEST
void waitreboot()
{
taskDelay (500);
reboot(BOOT_CLEAR);
}
#endif
